// SPDX-License-Identifier: MIT

pragma solidity ^0.6.12;
pragma experimental ABIEncoderV2;

import { SafeMath } from "@openzeppelin/contracts/math/SafeMath.sol";
import { LoopbackProxy } from "../v1/LoopbackProxy.sol";
import { AdminUpgradeableProxy } from "./AdminUpgradeableProxy.sol";
import { IERC20 } from "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import { Address } from "@openzeppelin/contracts/utils/Address.sol";

import { GovernancePatchUpgrade } from "./GovernancePatchUpgrade.sol";
import { TornadoStakingRewards } from "./TornadoStakingRewards.sol";
import { RelayerRegistry } from "./RelayerRegistry.sol";

/**
 * @notice Contract which should help the proposal deploy the necessary contracts.
 */
contract PatchProposalContractsFactory {
    /**
     * @notice Create a new TornadoStakingRewards contract.
     * @param governance The address of Tornado Cash Goveranance.
     * @param torn The torn token address.
     * @param registry The address of the relayer registry.
     * @return The address of the new staking contract.
     */
    function createStakingRewards(address governance, address torn, address registry)
        external
        returns (address)
    {
        return address(new TornadoStakingRewards(governance, torn, registry));
    }

    /**
     * @notice Create a new RelayerRegistry contract.
     * @param torn The torn token address.
     * @param governance The address of Tornado Cash Goveranance.
     * @param ens The ens registrar address.
     * @param staking The TornadoStakingRewards contract address.
     * @return The address of the new registry contract.
     */
    function createRegistryContract(
        address torn,
        address governance,
        address ens,
        address staking,
        address feeManager
    ) external returns (address) {
        return address(new RelayerRegistry(torn, governance, ens, staking, feeManager));
    }
}

/**
 * @notice Proposal which should patch governance against the metamorphic contract replacement vulnerability.
 */
contract PatchProposal {
    using SafeMath for uint256;
    using Address for address;

    address public immutable feeManagerProxyAddress = 0x5f6c97C6AD7bdd0AE7E0Dd4ca33A4ED3fDabD4D7;
    address public immutable registryProxyAddress = 0x58E8dCC13BE9780fC42E8723D8EaD4CF46943dF2;
    address public immutable ensAddress = 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e;

    IERC20 public constant TORN = IERC20(0x77777FeDdddFfC19Ff86DB637967013e6C6A116C);

    PatchProposalContractsFactory public immutable patchProposalContractsFactory;

    constructor(address _patchProposalContractsFactory) public {
        patchProposalContractsFactory = PatchProposalContractsFactory(_patchProposalContractsFactory);
    }

    /// @notice Function to execute the proposal.
    function executeProposal() external {
        // address(this) has to be governance
        address payable governance = payable(address(this));

        // Get the two contracts gov depends on
        address gasComp = address(GovernancePatchUpgrade(governance).gasCompensationVault());
        address vault = address(GovernancePatchUpgrade(governance).userVault());

        // Get the old staking contract
        TornadoStakingRewards oldStaking =
            TornadoStakingRewards(address(GovernancePatchUpgrade(governance).Staking()));

        // Get the small amount of TORN left
        oldStaking.withdrawTorn(TORN.balanceOf(address(oldStaking)));

        // And create a new staking logic contract
        TornadoStakingRewards newStakingImplementation = TornadoStakingRewards(
            patchProposalContractsFactory.createStakingRewards(
                address(governance), address(TORN), registryProxyAddress
            )
        );

        // Create new staking proxy contract (without initialization value)
        bytes memory empty;

        address newStaking =
            address(new AdminUpgradeableProxy(address(newStakingImplementation), address(governance), empty));

        // And a new registry implementation
        address newRegistryImplementationAddress = patchProposalContractsFactory.createRegistryContract(
            address(TORN), address(governance), ensAddress, newStaking, feeManagerProxyAddress
        );

        // Upgrade the registry proxy
        AdminUpgradeableProxy(payable(registryProxyAddress)).upgradeTo(newRegistryImplementationAddress);

        // Now upgrade the governance to the latest stuff
        LoopbackProxy(payable(governance)).upgradeTo(
            address(new GovernancePatchUpgrade(newStaking, gasComp, vault))
        );

        // Compensate TORN for staking
        TORN.transfer(newStaking, 94_092 ether);
    }
}
